package	jp.co.sds_corp.puzzle;

import java.awt.Point;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Vector;

import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;

public class NumPuzzle
{
	static final CmnLogger logger = CmnLogFactory.getLogger();
	Cell[][] cells, save;
	int		nMark;
	boolean	bCheck;
	boolean	bHint = false;
	boolean	stop = true;
	int		nHintX = -1, nHintY = -1, nHintNum = -1;
	NumPuzzlePnl parent;
	private Thread thr = null;

	public NumPuzzle() {
		cells = new Cell[9][9];
		for ( int i = 0; i < 9; i++ ) {
			for ( int j = 0; j < 9; j++ ) {
				cells[i][j] = new Cell();
			}
		}
		save = new Cell[9][9];
		for ( int i = 0; i < 9; i++ ) {
			for ( int j = 0; j < 9; j++ ) {
				save[i][j] = new Cell();
			}
		}
	}

	public NumPuzzle( String[] args ) throws Exception {
		this();
		if ( args.length != 0 ) {
			readFile( new FileInputStream( args[0] ));
		} else {
			inputQuestion();
		}
	}

	public NumPuzzle( NumPuzzlePnl parent ) {
		this();
		this.parent = parent;
	}

	public synchronized void readFile( InputStream inp ) throws Exception {
		BufferedReader br;
		String	s;
		int		x, y;

		logger.trace(this, "readFile");
		clearAll();
		try {
			br = new BufferedReader( new InputStreamReader( inp ));
			for ( y = 0; y < 9; y++ ) {
				s = br.readLine();
				if ( s == null ) {
					break;
				}
				for ( x = 0; x < 9; x++ ) {
					int data = Integer.valueOf( s.substring( x, x+1 ) ).intValue();
					cells[x][y].setNumber(data);
					if ( data != 0 ) {
						cells[x][y].setMode(Cell.Mode.QUESTION);
					}
				}
			}
			br.close();
		} catch ( IOException ex ) {
			logger.error(this, "file load error", ex);
			throw(ex);
		}
	}

	public void clear( int x, int y ) {
		cells[x][y].clear();
	}
	public void clearAll() {
		for ( int x = 0; x < 9; x++ ) {
			for ( int y = 0; y < 9; y++ ) {
				clear( x, y );
			}
		}
	}

	public void setData( int x, int y, int num, Cell.Mode mode) {
		cells[x][y].setNumber(num);
		cells[x][y].setMode(mode);
	}
	public int getData( int x, int y ) {
		return cells[x][y].getNumber();
	}
	public Cell.Mode getMode( int x, int y ) {
		return cells[x][y].getMode();
	}
	public void setAvailable( int x, int y, int num, boolean avail) {
		cells[x][y].setAvailable(num, avail);
	}

	void inputQuestion() throws Exception {
		int		x, y;
		byte[]	b = new byte[1024];

		logger.trace(this, "inputQuestion");
		clearAll();
		for ( y = 0; y < 9; y++ ) {
			for ( x = 0; x < 9; x++ ) {
				System.out.print( "(%d,%d)= "+  x + "," + y );
				try {
					System.in.read( b );
				} catch ( IOException ex ) {
					throw(ex);
				}
				int data = Integer.valueOf( new String( b ) ).intValue();
				cells[x][y].setNumber(data);
			}
		}
	}

	void printTable() {
		int		x, y;

		logger.debug( this, "|1|2|3|4|5|6|7|8|9|" );
		logger.debug( this, "+-+-+-+-+-+-+-+-+-+" );
		for ( y = 0; y < 9; y++ ) {
			StringBuilder buf = new StringBuilder();
			for ( x = 0; x < 9; x++ ) {
				buf.append( "|" + cells[x][y].getNumber() );
			}
			logger.debug( this, buf + "|" );
			logger.debug( this, "+-+-+-+-+-+-+-+-+-+" );
		}
/*
		for( int num = 1; num <= 9; num++ ) {
			logger.debug( "\nnum=" + num );
			logger.debug( "|1|2|3|4|5|6|7|8|9|" );
			logger.debug( "+-+-+-+-+-+-+-+-+-+" );
			for ( y = 0; y < 9; y++ ) {
				StringBuilder buf = new StringBuilder();
				for ( x = 0; x < 9; x++ ) {
					if (cells[x][y].getNumber() == num) {
						buf.append( "|" + num );
					} else {
						if( cells[x][y].available(num) ) {
							buf.append( "|O" );
						} else {
							buf.append( "|X" );
						}
					}
				}
				logger.debug( buf + "|" );
				logger.debug( "+-+-+-+-+-+-+-+-+-+" );
			}
		}
*/
	}

	public void copyAll() {
		int		x, y;

		setHint(false);
		setCheck(false);
		for ( y = 0; y < 9; y++ ) {
			for ( x = 0; x < 9; x++ ) {
				for ( int d = 1; d < 10; d++ ) {
					cells[x][y].setAvailable(d, true);
				}
			}
		}
		for ( y = 0; y < 9; y++ ) {
			for ( x = 0; x < 9; x++ ) {
				int num = cells[x][y].getNumber();
				if ( num != 0 ) {
					mark( num, x, y );
				}
			}
		}
	}

	void mark( int num, int x, int y ) {
		int	l, nx, ny;

		if ( isHint() ) {
			setHint(x, y, num);
			nMark++;
			return;
		}
		cells[x][y].setNumber(num);
		cells[x][y].setAvailable(num, false);
		nMark++;

		for ( l = 1; l <= 9; l++ ) {
			if ( l == num ) {
				continue;
			}
			cells[x][y].setAvailable(l, false);
		}
		for ( nx = 0; nx < 9; nx++ ) {
			if ( nx == x ) {
				continue;
			}
			cells[nx][y].setAvailable(num, false);
		}
		for ( ny = 0; ny < 9; ny++ ) {
			if ( ny == y ) {
				continue;
			}
			cells[x][ny].setAvailable(num, false);
		}
		markBlock( num, x/3, y/3 );
		if ( isCheck() && ( parent != null ))  {
			parent.setData();
			try {
				Thread.sleep( 100 );
			} catch ( InterruptedException ex ) {
				throw( new BreakException() );
			}
		}
	}

	void markBlock( int num, int bx, int by ) {
		int		x, y, nx, ny;

		for ( ny = 0; ny < 3; ny++ ) {
			y = by * 3 + ny;
			for ( nx = 0; nx < 3; nx ++ ) {
				x = bx * 3 + nx;
				if ( cells[x][y].getNumber() != num ) {
					cells[x][y].setAvailable(num, false);
				}
			}
		}
	}

	void check( int num ) {
		int		bx, by;

		checkHorizontal( num );
		checkVertical( num );

		for ( bx = 0; bx < 3; bx ++ ) {
			for ( by = 0; by < 3; by++ ) {
				checkBlock( num, bx, by );
			}
		}
	}

	void checkHorizontal( int num ) {
		int		x, y;
		int		pos;
		int		px = -1, py = -1;

		for ( y = 0; y < 9; y++ ) {
			pos = 0;
			for ( x = 0; x < 9; x ++ ) {
				if ( cells[x][y].available(num)) {
					pos++;
					px = x;
					py = y;
				}
			}
			if ( pos == 1 ) {
				mark( num, px, py );
			}
		}
	}

	void checkVertical( int num ) {
		int		x, y;
		int		pos;
		int		px = -1, py = -1;

		for ( x = 0; x < 9; x ++ ) {
			pos = 0;
			for ( y = 0; y < 9; y++ ) {
				if ( cells[x][y].available(num)) {
					pos++;
					px = x;
					py = y;
				}
			}
			if ( pos == 1 ) {
				mark( num, px, py );
			}
		}
	}

	void checkBlock( int num, int bx, int by ) {
		int		nx, ny, x, y;
		int		pos, px = -1, py = -1;

		pos = 0;
		for ( nx = 0; nx < 3; nx++ ) {
			x = bx * 3 + nx;
			for ( ny = 0; ny < 3; ny++ ) {
				y = by * 3 + ny;
				if ( cells[x][y].available(num)) {
					pos++;
					px = x;
					py = y;
				}
			}
		}
		if ( pos == 1 ) {
			mark( num, px, py );
		}
	}

	void checkOnly() {
		int		x, y;
		int		pos;
		int		pnum = -1;

		for ( x = 0; x < 9; x ++ ) {
			for ( y = 0; y < 9; y++ ) {
				pos = 0;
				for ( int num = 1; num < 10; num++ ) {
					if ( cells[x][y].available(num)) {
						pos++;
						pnum = num;
					}
				}
				if ( pos == 1 ) {
					mark( pnum, x, y );
				}
			}
		}
	}

	public int hint() {
		setHint(-1, -1, -1);

		copyAll();
		nMark = 0;
		setHint(true);

		errMsg = new StringBuffer();
		boolean bRet = checkLegal();
		if ( !bRet ) {
			putErrMsg( bRet );
			return -1;
		}

		for ( int num = 1; num < 10; num++ ) {
			for ( int bx = 0; bx < 3; bx ++ ) {
				for ( int by = 0; by < 3; by++ ) {
					checkBlock( num, bx, by );
					if ( nMark != 0 ) {
						return 3;
					}
				}
			}
			checkHorizontal( num );
			if ( nMark != 0 ) {
				return 1;
			}
			checkVertical( num );
			if ( nMark != 0 ) {
				return 2;
			}
		}
		checkOnly();
		if ( nMark != 0 ) {
			return 4;
		}
		return -1;
	}

	StringBuffer	errMsg;
	boolean checkLegal() {
		boolean	bRet;
		boolean bRes = true;
		for ( int x = 0; x < 9; x++ ) {
			for ( int y = 0; y < 9; y++ ) {
				int data = cells[x][y].getNumber();
				if ( data == 0 ) {
					bRet = isPossible( x, y );
					if ( !bRet ) {
						errMsg.append( "("+(x+1)+","+(y+1)+") "
										+ "にはどの数字も入りません\n" );
					}
				} else {
					bRet = isLegal( x, y, data );
					if ( !bRet ) {
						errMsg.append( "("+(x+1)+","+(y+1)+") "
										+ "は不正な数字が入っています\n" );
					}
				}
				if ( !bRet ) {
					bRes = false;
				}
			}
		}
		return bRes;
	}
	boolean checkNum() {
		boolean bRes = true;
		for ( int num = 1; num <= 9; num++ ) {
			int		nNum = 0;
			for ( int x = 0; x < 9; x++ ) {
				for ( int y = 0; y < 9; y++ ) {
					if ( cells[x][y].getNumber() == num ) {
						nNum++;
					}
				}
			}
			if ( nNum < 9 ) {
				errMsg.append( num + "が" + ( 9 - nNum ) + "個足りません\n" );
				bRes = false;
			} else
			if ( nNum > 9 ) {
				errMsg.append( num + "が" + ( nNum - 9 ) + "個多すぎます\n" );
				bRes = false;
			}
		}
		return bRes;
	}

	public boolean checkAll() {
		boolean	bRes = true;
		boolean	bRet;
		errMsg = new StringBuffer();
		bRet = checkLegal();
		if ( !bRet ) {
			bRes = false;
		}
		bRet = checkNum();
		if ( !bRet ) {
			bRes = false;
		}
		return bRes;
	}

	public boolean isInputFinished() {
		boolean finished = true;
		for ( int x = 0; x < 9; x++ ) {
			for ( int y = 0; y < 9; y++ ) {
				int data = cells[x][y].getNumber();
				if ( data == 0 ) {
					finished = false;
					break;
				}
			}
		}
		return finished;
	}

	boolean isPossible( int x, int y ) {
		boolean	bRet = false;
		for ( int d = 1; d <= 9; d++ ) {
			if ( cells[x][y].available(d)) {
				bRet = true;
				break;
			}
		}
		return bRet;
	}

	Vector<Point> vecErr = new Vector<Point>();
	public boolean isLegal( int x, int y, int num ) {
		boolean	bRet = true;
		vecErr.clear();
		for ( int i = 0; i < 9; i++ ) {
			if ( cells[i][y].getNumber() == num ) {
				if ( x != i ) {
					vecErr.add( new Point( i, y ));
					bRet = false;
				}
			}
		}
		for ( int i = 0; i < 9; i++ ) {
			if ( cells[x][i].getNumber() == num ) {
				if ( y != i ) {
					vecErr.add( new Point( x, i ));
					bRet = false;
				}
			}
		}
		int		tx, ty;
		for ( int ny = 0; ny < 3; ny++ ) {
			ty = (y/3) * 3 + ny;
			for ( int nx = 0; nx < 3; nx++ ) {
				tx = (x/3) * 3 + nx;
				if ( cells[tx][ty].getNumber() == num ) {
					if (( tx != x ) && ( ty != y )) {
						vecErr.add( new Point( nx, ny ));
						bRet = false;
					}
				}
			}
		}
		return bRet;
	}

	class MessageView implements Runnable {
		boolean bRet;
		MessageView(boolean bRet) {
			this.bRet = bRet;
		}

		@Override
		public void run() {
			if ( bRet ) {
				JOptionPane.showMessageDialog( parent,
							"完了しました",
							"Complete",
							JOptionPane.INFORMATION_MESSAGE);
			} else {
				JOptionPane.showMessageDialog( parent,
							errMsg.toString(),
							"Alert",
							JOptionPane.ERROR_MESSAGE);
			}
		}

	}
	void putErrMsg( boolean bRet ) {
		if (parent == null) {
			return;
		}
		MessageView msg = new MessageView(bRet);
		SwingUtilities.invokeLater(msg);
	}

	boolean auto() {
		boolean bRet;
		logger.trace(this, "auto");
		setHint(false);
		for ( ;; ) {
			nMark = 0;
			for ( int num = 1; num <= 9; num ++ ) {
				if (isStop()) {
					return true;
				}
				check( num );
			}
			checkOnly();
			logger.trace(this, "mark=" + nMark);
			if ( nMark == 0 ) {
				break;
			}
		}
		bRet = checkAll();
		logger.trace(this, "check=" + bRet);
		return bRet;
	}

	void copyData( Cell[][] dst, Cell[][] src ) {
		for ( int x = 0; x < 9; x++ ) {
			for ( int y = 0; y < 9; y++ ) {
				dst[x][y].copy(src[x][y]);
			}
		}
	}

	boolean trial( int num, int x, int y ) {
		boolean bRet, bRet2;
		logger.trace(this, "trial");
		mark( num, x, y );
		bRet = auto();
		if (isStop()) {
			copyData( cells, save );
			parent.setData();
			throw( new BreakException() );
		}
		parent.setData();
		try {
			Thread.sleep( 500 );
		} catch ( InterruptedException ex ) {
			copyData( cells, save );
			parent.setData();
			throw( new BreakException() );
		}
		if ( bRet ) {
			return true;
		}
		bRet2 = checkLegal();
		if ( !bRet2 ) {
			save[x][y].setAvailable(num, false);
		}
		copyData( cells, save );
		parent.setData();
		if (isStop()) {
			throw( new BreakException() );
		}
		try {
			Thread.sleep( 500 );
		} catch ( InterruptedException ex ) {
			copyData( cells, save );
			parent.setData();
			throw( new BreakException() );
		}
		return false;
	}

	class Resolver implements Runnable {
		public void run() {
			int num=0, x=0, y=0;
			boolean bTry = false;

			copyAll();
//			if ( parent == null ) {
				logger.debug( this, "Q" );
				printTable();
//			}

			try {
				boolean bRet;
				setCheck(true);
				bRet = auto();
				if (isStop()) {
					return;
				}
				if ( !bRet && parent.bAssist ) {
					bTry = true;
					setCheck(false);
					copyData( save, cells );
					for ( x = 0; x < 9; x++ ) {
						for ( y = 0; y < 9; y++ ) {
							if ( save[x][y].getNumber() == 0 ) {
								for ( num = 1; num < 10; num++ ) {
									if ( save[x][y].available(num)) {
										bRet = trial( num, x, y );
										if ( bRet ) {
											break;
										}
									}
								}
								if ( bRet ) {
									break;
								}
							}
						}
						if ( bRet ) {
							break;
						}
					}
					copyData( cells, save );
					parent.setData();
				}
				if ( bRet ) {
					if ( bTry ) {
						copyData( cells, save );
						parent.setData();
						try {
							Thread.sleep( 1000 );
						} catch ( InterruptedException ex ) {
							setStop(true);
							return;
						}
						if (isStop()) {
							return;
						}
						setCheck(true);
						mark( num, x, y );
						auto();
						if (isStop()) {
							return;
						}
					}
				}
//				if ( parent == null ) {
					logger.debug( this, "A" );
					printTable();
//				}
				if (parent != null) {
					parent.BT_AUTO.setText( "Auto" );
					parent.setAuto(false);
					parent.setData();
				}
				putErrMsg( bRet );
			} catch ( BreakException ex ) {
			}
			setStop(true);
		}
	}

	public synchronized boolean isCheck() {
		return bCheck;
	}

	public synchronized void setCheck(boolean bCheck) {
		this.bCheck = bCheck;
	}

	public synchronized boolean isHint() {
		return bHint;
	}

	public synchronized void setHint(boolean bHint) {
		this.bHint = bHint;
	}

	public synchronized void setHint(int x, int y, int num) {
		nHintX = x;
		nHintY = y;
		nHintNum = num;
	}

	public synchronized int getHintX() {
		return nHintX;
	}
	public synchronized int getHintY() {
		return nHintY;
	}
	public synchronized int getHintNum() {
		return nHintNum;
	}

	public synchronized void setStop(boolean var) {
		this.stop = var;
	}
	public synchronized boolean isStop() {
		return this.stop;
	}

	public String getErrMsg() {
		return errMsg.toString();
	}

	public void start() {
		if ( isStop() ) {
			setStop(false);
			thr = new Thread(new Resolver());
			thr.start();
		}
	}

	public void stopThread() {
		if ( thr != null ) {
			setStop(true);
			waitThread();
		}
	}

	void waitThread() {
		try {
			thr.join();
		} catch (InterruptedException ex) {
		}
	}

	public static void main( String[] args ) {
		try {
			NumPuzzle obj = new NumPuzzle( args );
			obj.start();
		} catch (Exception ex) {
			logger.error(NumPuzzle.class, "start error", ex);
			System.exit(9);
		}
	}
}
